#!/usr/bin/env python3
# coding=utf-8

import subprocess
import re

def get_nvlink_info():
    # 运行 nvidia-smi nvlink -R 命令并获取输出
    result = subprocess.run(['nvidia-smi', 'nvlink', '-R'], stdout=subprocess.PIPE, text=True, check=True)
    output = result.stdout.strip()
    return output

def parse_nvlink_info(output):
    gpu_nvlink_count = {}
    gpu_nvlink_status = {}
    gpu_uuids = {}  # 存储每个 GPU 的 UUID
    
    # 调整正则表达式以匹配每个GPU的NVLink信息
    gpu_pattern = re.compile(r'GPU\s+(\d+):\s+(NVIDIA.*?)\s+\(UUID:\s+GPU-([a-f0-9\-]+)\)((?:\n\s+Link \d+: Remote Device [0-9a-fA-F:.]+: Link \d+)+)', re.MULTILINE)
    link_pattern = re.compile(r'Link \d+: Remote Device ([0-9a-fA-F:.]+): Link \d+', re.MULTILINE)
    
    matches = gpu_pattern.findall(output)
    
    for match in matches:
        gpu_index = match[0]
        gpu_uuid = match[2]
        links_info = match[3]
        
         # 存储 GPU UUID
        gpu_uuids[gpu_index] = gpu_uuid

        # 统计NVLink数量
        if links_info:
            num_nvlinks = len(link_pattern.findall(links_info))
            all_links_valid = all(link != '00000000:00:00.0' for link in link_pattern.findall(links_info))
        else:
            num_nvlinks = 0
            all_links_valid = False
        
        # 更新GPU的NVLink数量和状态
        gpu_nvlink_count[gpu_index] = num_nvlinks
        gpu_nvlink_status[gpu_index] = num_nvlinks == 18 and all_links_valid
    
    return gpu_nvlink_count, gpu_nvlink_status, gpu_uuids

def get_gpu_info():
    # 运行 nvidia-smi -L 命令并获取输出
    result = subprocess.run(['nvidia-smi', '-L'], stdout=subprocess.PIPE, text=True, check=True)
    output = result.stdout.strip()
    return output

def count_gpus(output):
    # 使用正则表达式匹配并统计 GPU 信息行
    gpu_pattern = re.compile(r'^GPU \d+:', re.MULTILINE)
    gpu_count = len(gpu_pattern.findall(output))
    return gpu_count

def get_gpu_bus_ids():
    try:
        # 运行 nvidia-smi 命令并捕获输出
        result = subprocess.run(['nvidia-smi'], stdout=subprocess.PIPE, text=True, check=True)
        output = result.stdout

        # 使用正则表达式匹配并提取 Bus-Id
        bus_ids = re.findall(r'00000000:([0-9A-Fa-f]{2}:[0-9A-Fa-f]{2}\.[0-9A-Fa-f])', output)

        return bus_ids

    except subprocess.CalledProcessError as e:
        print(f"Error: Failed to run command. {e}")
        return None

def check_pci_status(bus_id):
    try:
        # 构建命令
        command = f"lspci -s {bus_id} -vvv | grep LnkSta:"
        # 运行命令并获取输出
        result = subprocess.run(command, shell=True, stdout=subprocess.PIPE, text=True, check=True)
        output = result.stdout
        
        # 使用正则表达式匹配 Speed 和 Width 信息
        match = re.search(r'Speed\s+32GT/s\s+\(ok\),\s+Width\s+x16\s+\(ok\)', output)

        # 如果匹配成功，返回 1，否则返回 0
        if match:
            return 1
        else:
            return 0
    except subprocess.CalledProcessError as e:
        print(f"Error: Failed to run command. {e}")
        return 0

def check_system_status(count_normal, count_abnormal):
    if count_normal == 8 and count_abnormal == 0:
        return 1
    else:
        return 0

def generate_prometheus_format(gpu_nvlink_count, gpu_nvlink_status, gpu_uuids, gpu_count, input_name, region, azone, product):
    # 生成 Prometheus 格式的文本数据
    prometheus_data = ""
    
    # 添加 nvlink_count_gpu 的帮助信息和类型
    prometheus_data += (
        f"# HELP {input_name}nvlink_count_gpu Number of NVLinks for each GPU\n"
        f"# TYPE {input_name}nvlink_count_gpu gauge\n"
    )

    # 每个GPU的NVLink数量指标
    for gpu_index, count in gpu_nvlink_count.items():
        gpu_uuid = gpu_uuids.get(gpu_index, 'unknown_uuid')  # 根据索引获取实际的 GPU UUID
        
        prometheus_data += (
            f'{input_name}nvlink_count_gpu{{region="{region}", azone="{azone}", product="{product}", gpu_index="{gpu_index}", gpu_uuid="{gpu_uuid}"}} {count}\n'
        )
    
    # 添加 nvlink_status_gpu 的帮助信息和类型
    prometheus_data += (
        f"# HELP {input_name}nvlink_status_gpu NVLink status for each GPU\n"
        f"# TYPE {input_name}nvlink_status_gpu gauge\n"
    )
    
    # 每个GPU的NVLink状态指标
    for gpu_index, status in gpu_nvlink_status.items():
        gpu_uuid = gpu_uuids.get(gpu_index, 'unknown_uuid')  # 根据索引获取实际的 GPU UUID
        prometheus_data += (
            f'{input_name}nvlink_status_gpu{{region="{region}", azone="{azone}", product="{product}", gpu_index="{gpu_index}", gpu_uuid="{gpu_uuid}"}} {int(status)}\n'
        )
    
    # 整体服务器GPU NVLink状态指标
    overall_status = all(gpu_nvlink_status.values()) and all(count == 18 for count in gpu_nvlink_count.values())
    prometheus_data += (
        f"# HELP {input_name}overall_nvlink_status Overall NVLink status for the server\n"
        f"# TYPE {input_name}overall_nvlink_status gauge\n"
        f'{input_name}overall_nvlink_status{{region="{region}", azone="{azone}", product="{product}"}} {int(overall_status)}\n'
    )
    
    # GPU数量指标
    prometheus_data += (
        f"# HELP {input_name}gpu_count Number of GPUs in the server\n"
        f"# TYPE {input_name}gpu_count gauge\n"
        f'{input_name}gpu_count{{region="{region}", azone="{azone}", product="{product}"}} {gpu_count}\n'
    )
    
    # 调用函数获取 GPU 的 Bus-Id 信息
    gpu_bus_ids = get_gpu_bus_ids()

    # 检查是否成功获取到 Bus-Id
    if gpu_bus_ids is None:
        print("Failed to retrieve GPU Bus-Ids.")
        return prometheus_data  # 直接返回已生成的部分数据

    # 初始化计数器
    count_normal = 0
    count_abnormal = 0

    # 检查每个 Bus-Id 的状态并更新计数器
    for bus_id in gpu_bus_ids:
        status = check_pci_status(bus_id)
        # print(f"Bus-Id {bus_id}: Status {status}")
        if status == 1:
            count_normal += 1
        else:
            count_abnormal += 1

    # 打印统计结果
    # print(f"\n统计结果: 返回值为 1 的数量: {count_normal}, 返回值为 0 的数量: {count_abnormal}")

    # 检查系统状态
    pcie_status = check_system_status(count_normal, count_abnormal)

    # PCIe 状态指标
    prometheus_data += (
        f"# HELP {input_name}pcie_status PCIE Status for GPUs\n"
        f"# TYPE {input_name}pcie_status gauge\n"
        f'{input_name}pcie_status{{region="{region}", azone="{azone}", product="{product}", count_normal="{count_normal}", count_abnormal="{count_abnormal}"}} {pcie_status}\n'
    )

    # 判断并生成 GPU 总体状态汇总指标
    gpu_overall_status = 1 if gpu_count == 8 and overall_status == 1 and pcie_status == 1 else 0
    prometheus_data += (
        f"# HELP {input_name}gpu_overall_status GPU Overall Status\n"
        f"# TYPE {input_name}gpu_overall_status gauge\n"
        f'{input_name}gpu_overall_status{{region="{region}", azone="{azone}", product="{product}"}} {gpu_overall_status}\n'
    )

    return prometheus_data


if __name__ == "__main__":
    # 自定义监控指标和标签
    input_name = "node_"
    region = "beijing"
    azone = "datacenter01"
    product = "prod"
    
    try:
        # 获取NVLink命令输出
        nvlink_output = get_nvlink_info()
        
        # 解析NVLink信息
        gpu_nvlink_count, gpu_nvlink_status, gpu_uuids = parse_nvlink_info(nvlink_output)
        
        # 获取GPU命令输出
        gpu_output = get_gpu_info()
        
        # 统计GPU数量
        gpu_count = count_gpus(gpu_output)
        
        # 生成Prometheus格式的数据
        prometheus_data = generate_prometheus_format(gpu_nvlink_count, gpu_nvlink_status, gpu_uuids, gpu_count, input_name, region, azone, product)
        
        # 打印Prometheus格式的数据
        print(prometheus_data)
        
    except subprocess.CalledProcessError as e:
        print(f"Command failed with error: {e}")

