#!/usr/bin/env python3
# coding=utf-8

import subprocess

def get_ubuntu_version():
    try:
        result = subprocess.run(['lsb_release', '-d'], check=True, stdout=subprocess.PIPE, text=True)
        version_info = result.stdout.strip().split('\t')[1]
        return version_info
    except subprocess.CalledProcessError as e:
        print(f"Error fetching Ubuntu version: {e}")
        return None

def get_kernel_version():
    try:
        result = subprocess.run(['uname', '-r'], check=True, stdout=subprocess.PIPE, text=True)
        kernel_version = result.stdout.strip()
        return kernel_version
    except subprocess.CalledProcessError as e:
        print(f"Error fetching kernel version: {e}")
        return None

def get_ofed_version():
    try:
        result = subprocess.run(['ofed_info', '-s'], check=True, stdout=subprocess.PIPE, text=True)
        ofed_version = result.stdout.strip().split(':')[0]
        return ofed_version
    except subprocess.CalledProcessError as e:
        print(f"Error fetching OFED version: {e}")
        return None

def get_firmware_version():
    try:
        result = subprocess.run(['ibv_devinfo', '|', 'grep', 'fw_ver'], check=True, stdout=subprocess.PIPE, text=True, shell=True)
        lines = result.stdout.strip().split('\n')
        firmware_versions = []
        for line in lines:
            if line.strip().startswith("fw_ver:"):
                firmware_version = line.strip().split()[1]
                firmware_versions.append(firmware_version)
        return firmware_versions
    except subprocess.CalledProcessError as e:
        print(f"Error fetching firmware version: {e}")
        return None

def check_firmware_versions(firmware_versions):
    return all(version == '28.39.3004' for version in firmware_versions)

def generate_prometheus_metric(metric_name, metric_help, is_correct_version):
    metric_name = f"node_{metric_name}"  # Adding 'node_' prefix
    metric_value = 1 if is_correct_version else 0

    prometheus_data = f"# HELP {metric_name} {metric_help}\n"
    prometheus_data += f"# TYPE {metric_name} gauge\n"
    prometheus_data += f"{metric_name} {metric_value}\n"

    return prometheus_data

def generate_version_overall_metric(metrics):
    overall_value = 1 if all(metrics.values()) else 0

    prometheus_data = "# HELP node_version_overall_status Boolean metric to indicate overall node version status\n"
    prometheus_data += "# TYPE node_version_overall_status gauge\n"
    prometheus_data += f"node_version_overall_status {overall_value}\n"

    return prometheus_data

if __name__ == "__main__":
    # Check Ubuntu version
    ubuntu_version = get_ubuntu_version()
    is_correct_ubuntu_version = ubuntu_version == 'Ubuntu 22.04.4 LTS' if ubuntu_version else False

    # Check kernel version
    kernel_version = get_kernel_version()
    is_correct_kernel_version = kernel_version == '5.15.0-105-generic' if kernel_version else False

    # Check OFED version
    ofed_version = get_ofed_version()
    is_correct_ofed_version = ofed_version == 'MLNX_OFED_LINUX-23.10-2.1.3.1' if ofed_version else False

    # Check firmware version
    firmware_versions = get_firmware_version()
    is_correct_firmware_version = check_firmware_versions(firmware_versions) if firmware_versions else False

    # Print results
    #if ubuntu_version:
    #    print(f"Ubuntu version: {ubuntu_version}")
    #    print(f"Is correct Ubuntu version: {is_correct_ubuntu_version}")
    #else:
    #    print("Failed to detect Ubuntu version.")

    #if kernel_version:
    #    print(f"Kernel version: {kernel_version}")
    #    print(f"Is correct kernel version: {is_correct_kernel_version}")
    #else:
    #    print("Failed to detect kernel version.")

    #if ofed_version:
    #    print(f"OFED version: {ofed_version}")
    #    print(f"Is correct OFED version: {is_correct_ofed_version}")
    #else:
    #    print("Failed to detect OFED version.")

    #if firmware_versions:
    #    print(f"Firmware versions: {', '.join(firmware_versions)}")
    #    print(f"Is correct firmware version (28.39.3004): {is_correct_firmware_version}")
    #else:
    #    print("Failed to detect firmware versions.")

    # Generate Prometheus metrics
    prometheus_metrics = ""
    prometheus_metrics += generate_prometheus_metric("ubuntu_version_check", "Boolean metric to check if the Ubuntu version is 22.04.4", is_correct_ubuntu_version)
    prometheus_metrics += generate_prometheus_metric("kernel_version_check", "Boolean metric to check if the kernel version is 5.15.0-105-generic", is_correct_kernel_version)
    prometheus_metrics += generate_prometheus_metric("ofed_version_check", "Boolean metric to check if the OFED version is MLNX_OFED_LINUX-23.10-2.1.3.1", is_correct_ofed_version)
    prometheus_metrics += generate_prometheus_metric("firmware_version_check", "Boolean metric to check if all firmware versions are 28.39.3004", is_correct_firmware_version)

    # Generate overall version Prometheus metric
    metrics = {
        "ubuntu_version_check": is_correct_ubuntu_version,
        "kernel_version_check": is_correct_kernel_version,
        "ofed_version_check": is_correct_ofed_version,
        "firmware_version_check": is_correct_firmware_version
    }
    prometheus_metrics += generate_version_overall_metric(metrics)

    print(prometheus_metrics)

