#!/usr/bin/env python3
# coding=utf-8

import subprocess
import re
import logging

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

# Function to get MST status
def get_mst_status():
    try:
        result = subprocess.run(['mst', 'status', '-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        if result.returncode != 0:
            logging.error("Error executing mst status -v: %s", result.stderr)
            return None
        return result.stdout
    except Exception as e:
        logging.exception("Exception occurred while executing mst status -v")
        return None

# Function to parse MST output and check RDMA device status
def parse_mst_output(output):
    device_lines = re.findall(r'ConnectX[57]\(rev:\d+\)\s+\S+\s+\S+\s+(mlx5_\d+)\s+net-\S+\s+\d+', output)
    rdma_devices = [line.split()[0] for line in device_lines]
    compute_plane_nics = [f'mlx5_{i:03d}' for i in range(101, 109)]
    storage_plane_nic = 'mlx5_201'
    nic_status = {nic: (nic in rdma_devices) for nic in compute_plane_nics + [storage_plane_nic]}
    overall_status = all(nic_status.values())
    return nic_status, overall_status

# Function to generate Prometheus format output for MST status
def generate_prometheus_format_mst(nic_status, overall_status, region, azone, product, input_name):
    prometheus_output = []
    labels = f'region="{region}", azone="{azone}", product="{product}"'
    prometheus_output.append(f'# HELP {input_name}nic_contain_status NIC contain status (1 if present, 0 if absent)')
    prometheus_output.append(f'# TYPE {input_name}nic_contain_status gauge')

    for nic, status in nic_status.items():
        prometheus_output.append(f'{input_name}nic_contain_status{{nic_name="{nic}", {labels}}} {int(status)}')

    prometheus_output.append(f'# HELP {input_name}overall_nic_contain_status Overall NIC contain status (1 if all present, 0 if any absent)')
    prometheus_output.append(f'# TYPE {input_name}overall_nic_contain_status gauge')
    prometheus_output.append(f'{input_name}overall_nic_contain_status{{{labels}}} {int(overall_status)}')

    return '\n'.join(prometheus_output)

# Function to get topology matrix using nvidia-smi topo --matrix
def get_topo_matrix():
    try:
        result = subprocess.run(['nvidia-smi', 'topo', '--matrix'], capture_output=True, text=True, check=True)
        if result.returncode != 0:
            raise RuntimeError("Error executing nvidia-smi topo --matrix")
        return result.stdout
    except subprocess.CalledProcessError as e:
        logging.error(f"Error executing nvidia-smi topo --matrix: {e}")
        return None
    except RuntimeError as e:
        logging.error(e)
        return None

# Function to check required NICs for GPU connections
def check_required_nics(output):
    lines = output.splitlines()
    nic_legend = {}
    nic_legend_start = False
    for line in lines:
        if 'NIC Legend:' in line:
            nic_legend_start = True
            continue
        if nic_legend_start:
            if not line.strip():
                continue
            if any(keyword in line for keyword in ['GPU', 'Legend']):
                break
            match = re.match(r'\s*(NIC\d{1,2}):\s*(mlx5_\d{3})', line)
            if match:
                nic_legend[match.group(1)] = match.group(2)

    required_nics = {f'mlx5_{i:03}' for i in range(101, 109)}
    required_nics.add('mlx5_201')
    missing_nics = required_nics - set(nic_legend.values())
    return len(missing_nics) == 0, nic_legend

# Function to check PIX connections between GPUs and NICs
def check_pix_connections(output):
    lines = output.splitlines()
    gpu_nic_count = 0
    gpu_indices = []
    nic_indices = []
    pix_status = {}
    for line in lines:
        if line.startswith('GPU') or line.startswith('NIC'):
            gpu_nic_count += 1

    matrix = []
    matrix_start = False
    for line in lines:
        if 'Legend:' in line.strip():
            break
        if matrix_start:
            if line.strip():
                matrix.append(line.split()[:gpu_nic_count])
        if 'GPU' in line and 'NIC' in line:
            matrix_start = True

    if not matrix:
        raise ValueError("Failed to parse matrix section")

    matrix_t = list(map(list, zip(*matrix)))
    gpu_indices = [i for i, col in enumerate(matrix_t[0]) if col.startswith('GPU')]
    nic_indices = [i for i, col in enumerate(matrix_t[0]) if col.startswith('NIC')]

    for gi in gpu_indices:
        for ni in nic_indices:
            if matrix_t[ni][gi] == 'PIX':
                gpu_id = f"GPU{gi}"
                nic_id = f"NIC{ni - gi}"
                pix_status[gpu_id] = 1

    return pix_status

# Function to generate Prometheus format output for topology matrix
def generate_prometheus_format_topo(topo_output, region, azone, product, input_name):
    prometheus_data = ""
    nic_satisfied, nic_legend = check_required_nics(topo_output)
    nic_satisfied_value = 1 if nic_satisfied else 0

    prometheus_data += (
        f"# HELP {input_name}nic_satisfied NIC Satisfied for GPUs\n"
        f"# TYPE {input_name}nic_satisfied gauge\n"
    )

    for nic_id in nic_legend.keys():
        prometheus_data += (
            f'{input_name}nic_satisfied{{region="{region}", azone="{azone}", product="{product}", nic_id="{nic_id}"}} {nic_satisfied_value}\n'
        )

    pix_status = check_pix_connections(topo_output)
    pix_satisfied_value = 1 if len(pix_status) == 8 else 0

    prometheus_data += (
        f"# HELP {input_name}gpu_pix_status GPU PIX Status\n"
        f"# TYPE {input_name}gpu_pix_status gauge\n"
    )

    for gpu_id, status in pix_status.items():
        prometheus_data += (
            f'{input_name}gpu_pix_status{{region="{region}", azone="{azone}", product="{product}", gpu_id="{gpu_id}"}} {status}\n'
        )

    total_nic_satisfied = len(nic_legend) == 9 and nic_satisfied_value == 1
    overall_status = int(total_nic_satisfied and pix_satisfied_value == 1)

    prometheus_data += (
        f"# HELP {input_name}overall_pix_status Overall Status of GPU and NIC connections\n"
        f"# TYPE {input_name}overall_pix_status gauge\n"
        f'{input_name}overall_pix_status{{region="{region}", azone="{azone}", product="{product}"}} {overall_status}'
    )

    return prometheus_data, overall_status

# Function to parse MST output and extract PCI addresses
def parse_pci_addresses(mst_output):
    pci_addresses = []
    lines = mst_output.splitlines()
    for line in lines:
        match = re.search(r'ConnectX[57]\(rev:0\)\s+\S+\s+([0-9a-f]{2}:[0-9a-f]{2}\.[0-9a-f]{1})\s+mlx5_\d{3}\s+net-\w+\d+\s+\d{1}\s{5}', line)
        if match:
            pci_addresses.append(match.group(1))
    return pci_addresses

# Function to execute lspci command and get PCI link status
def get_link_status(pci_address):
    try:
        cmd = f'lspci -s {pci_address} -vvv | grep LnkSta:'
        result = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        if result.returncode != 0:
            logging.error("Error executing lspci for %s: %s", pci_address, result.stderr)
            return None
        return result.stdout.strip()
    except Exception as e:
        logging.exception("Exception occurred for %s", pci_address)
        return None

def check_link_status(link_status):
    speed_match = re.search(r'Speed (\d+GT/s)', link_status)
    width_match = re.search(r'Width (x\d+)', link_status)
    
    speed = speed_match.group(1) if speed_match else "unknown"
    width = width_match.group(1) if width_match else "unknown"
    
    is_ok = "Speed 32GT/s (ok)" in link_status and "Width x16 (ok)" in link_status
    
    return is_ok, speed, width

# Function to check PCI link status and generate Prometheus format output
def generate_prometheus_format_pci(pci_addresses, input_name, region, azone, product):
    metrics = []
    metrics.append(f'# HELP {input_name}gpu_link_status Link status of GPU PCI devices')
    metrics.append(f'# TYPE {input_name}gpu_link_status gauge')
    overall_status = 1

    for pci_address in pci_addresses:
        try:
            link_status = get_link_status(pci_address)
            if link_status:
                is_ok, speed, width = check_link_status(link_status)
                status_value = 1 if is_ok else 0
                metrics.append(
                    f'{input_name}gpu_link_status{{pci="{pci_address}",speed="{speed}",width="{width}",region="{region}",azone="{azone}",product="{product}"}} {status_value}'
                )
                if not is_ok:
                    overall_status = 0
            else:
                metrics.append(
                    f'{input_name}gpu_link_status{{pci="{pci_address}",speed="unknown",width="unknown",region="{region}",azone="{azone}",product="{product}"}} 0'
                )
                overall_status = 0
        except Exception as e:
            logging.error("Exception occurred while processing %s: %s", pci_address, e)
            metrics.append(
                f'{input_name}gpu_link_status{{pci="{pci_address}",speed="unknown",width="unknown",region="{region}",azone="{azone}",product="{product}"}} 0'
            )
            overall_status = 0

    metrics.append(f'# HELP {input_name}gpu_overall_link_status Overall link status of GPU nodes')
    metrics.append(f'# TYPE {input_name}gpu_overall_link_status gauge')
    metrics.append(f'{input_name}gpu_overall_link_status{{region="{region}",azone="{azone}",product="{product}"}} {overall_status}')
    return '\n'.join(metrics), overall_status

# Function to execute ibstatus command and get InfiniBand device status
def get_ibstatus_output():
    try:
        result = subprocess.run(['ibstatus'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding='utf-8')
        if result.returncode != 0:
            logging.error("Error executing ibstatus")
            logging.error(result.stderr)
            return None
        return result.stdout
    except Exception as e:
        logging.exception("Exception occurred: %s", e)
        return None

# Function to parse ibstatus output and extract device information
def parse_ibstatus(output):
    devices_info = {}
    pattern = r"Infiniband device '([^']+)' port 1 status:\s+" \
              r"default gid:\s+(\S+)\s+" \
              r"base lid:\s+(0x[0-9a-fA-F]+)\s+" \
              r"sm lid:\s+(0x[0-9a-fA-F]+)\s+" \
              r"state:\s+(\d+):\s+(\w+)\s+" \
              r"phys state:\s+(\d+):\s+(\w+)\s+" \
              r"rate:\s+([\d\s]+ Gb/sec) \(4X NDR\)\s+" \
              r"link_layer:\s+(InfiniBand)"

    matches = re.finditer(pattern, output, re.MULTILINE)

    for match in matches:
        device_name = match.group(1)
        default_gid = match.group(2)
        base_lid = match.group(3)
        sm_lid = match.group(4)
        state_num = match.group(5)
        phys_state_num = match.group(7)
        rate = match.group(9)

        devices_info[device_name] = {
            'default_gid': default_gid,
            'base_lid': base_lid,
            'sm_lid': sm_lid,
            'state': state_num,
            'phys_state': phys_state_num,
            'rate': rate,
            'link_layer': 'InfiniBand'
        }

    return devices_info

# Function to calculate boolean values and overall status for InfiniBand devices
def calculate_boolean_values(devices_info):
    device_bool_values = {}
    all_states_bool = []
    overall_status = 1

    for device, info in devices_info.items():
        state_ok = info['state'] == '4'
        phys_state_ok = info['phys_state'] == '5'
        rate_ok = info['rate'] == '400 Gb/sec'
        link_layer_ok = info['link_layer'] == 'InfiniBand'

        device_status = int(state_ok and phys_state_ok and rate_ok and link_layer_ok)
        device_bool_values[device] = device_status
        all_states_bool.append(device_status)

        if device_status == 0:
            overall_status = 0

    return device_bool_values, all_states_bool, overall_status

# Function to generate Prometheus format output for InfiniBand devices
def generate_prometheus_output_ib(devices_info, all_states_bool, overall_status, input_name="node_", region="beijing", azone="datacenter01", product="prod"):
    prom_output = []
    prom_output.append(f"# HELP {input_name}ib_device_status Indicates whether each IB device meets criteria (1 = OK, 0 = Not OK)")
    prom_output.append("# TYPE ib_device_status gauge")

    for device, info in devices_info.items():
        prom_output.append(
            f'{input_name}ib_device_status{{region="{region}", azone="{azone}", product="{product}", '
            f'device="{device}", state="{info["state"]}", phys_state="{info["phys_state"]}", rate="{info["rate"]}", '
            f'link_layer="{info["link_layer"]}"}} 1'
        )

    prom_output.append(f"# HELP {input_name}ib_devices_overall_status Indicates overall status of all IB devices (1 = OK, 0 = Not OK)")
    prom_output.append("# TYPE ib_devices_overall_status gauge")
    prom_output.append(f'{input_name}ib_devices_overall_status{{region="{region}", azone="{azone}", product="{product}"}} {overall_status}')

    return "\n".join(prom_output)

if __name__ == "__main__":
    input_name = "node_"
    region = "beijing"
    azone = "datacenter01"
    product = "prod"

    try:
        # MST Status Section
        mst_output = get_mst_status()
        if mst_output:
            nic_status, overall_status_mst = parse_mst_output(mst_output)
            prometheus_metrics_mst = generate_prometheus_format_mst(nic_status, overall_status_mst, region, azone, product, input_name)
            print(prometheus_metrics_mst)
        else:
            print("Failed to get mst status output.")

        # Topology Matrix Section
        topo_output = get_topo_matrix()
        if topo_output:
            prometheus_metrics_topo, overall_pix_status = generate_prometheus_format_topo(topo_output, region, azone, product, input_name)
            print(prometheus_metrics_topo)
        else:
            print("Failed to get topology matrix output.")

        # PCI Link Status Section
        if mst_output:
            pci_addresses = parse_pci_addresses(mst_output)
            prometheus_metrics_pci, pci_overall_status = generate_prometheus_format_pci(pci_addresses, input_name, region, azone, product)
            print(prometheus_metrics_pci)
        else:
            print("Failed to get PCI addresses.")

        # IB Status Section
        ibstatus_output = get_ibstatus_output()
        if ibstatus_output:
            devices_info = parse_ibstatus(ibstatus_output)
            device_bool_values, all_states_bool, overall_status_ib = calculate_boolean_values(devices_info)
            prometheus_metrics_ib = generate_prometheus_output_ib(devices_info, all_states_bool, overall_status_ib, input_name, region, azone, product)
            print(prometheus_metrics_ib)
        else:
            print("Failed to get ibstatus output.")

        # Overall Status Section
        overall_status = int(overall_status_mst and overall_pix_status and overall_status_ib and pci_overall_status)
        prometheus_output = (
            f"# HELP {input_name}network_card_overall_status Overall Status of all components\n"
            f"# TYPE {input_name}network_card_overall_status gauge\n"
            f'{input_name}network_card_overall_status{{region="{region}", azone="{azone}", product="{product}"}} {overall_status}'
        )
        print(prometheus_output)

    except subprocess.CalledProcessError as e:
        print(f"Command failed with error: {e}")


